<?php

declare(strict_types=1);

namespace Drupal\tfa_smsapi\Plugin\TfaSetup;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\tfa\Plugin\TfaSetupInterface;
use Drupal\tfa_smsapi\Plugin\TfaValidation\TfaSmsapiValidation;

/**
 * TfaSmsapi setup class to setup SMSAPI validation.
 *
 * @TfaSetup(
 *   id = "tfa_smsapi_setup",
 *   label = @Translation("TFA SMSAPI Setup"),
 *   description = @Translation("TFA SMSAPI Setup Plugin"),
 *   setupMessages = {
 *    "saved" = @Translation("SMSAPI code verified."),
 *    "skipped" = @Translation("SMSAPI codes not enabled.")
 *   }
 * )
 */
class TfaSmsapiSetup extends TfaSmsapiValidation implements TfaSetupInterface {

  /**
   * {@inheritdoc}
   */
  public function getSetupForm(array $form, FormStateInterface $form_state) {
    $form = $this->getForm($form, $form_state);

    $phone = $this->getUserData('tfa', 'tfa_smsapi_phone', $this->uid, $this->userData);

    if (!empty($phone)) {
      [$countryCode, $phone] = explode('-', $phone);
    }
    else {
      $phone = $countryCode = '';
    }

    $form['country_code'] = [
      '#type' => 'number',
      '#title' => $this->t('Your country code number.'),
      '#required' => TRUE,
      '#min' => 1,
      '#step' => 1,
      '#max' => 999,
      '#description' => $this->t('Please enter your country code number, eg. 48 for Poland'),
      '#default_value' => $countryCode,
    ];

    $form['phone'] = [
      '#type' => 'tel',
      '#title' => $this->t('Your phone number'),
      '#required' => TRUE,
      '#attributes' => [
        'pattern' => '^\d{9,11}$',
      ],
      '#default_value' => $phone,
      '#description' => $this->t('Please enter your telephone number (numbers only) eg. 567665454'),
    ];

    $form['actions']['login']['#value'] = $this->t('Verify and save');
    $form['code']['#description'] = $this->t("A verification code will be generated and sent to your phone after you provide your phone number and click the button 'Get Verification code'. Enter it below and click 'Verify and save' to finish setting up this two factor authentication.");

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateSetupForm(array $form, FormStateInterface $form_state) {
    return $this->validateForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitSetupForm(array $form, FormStateInterface $form_state) {
    $phone = $form_state->getValue('country_code') . '-' . $form_state->getValue('phone');
    $this->setUserData('tfa', ['tfa_smsapi_phone' => $phone], $this->uid, $this->userData);
    $this->setUserData('tfa', ['tfa_smsapi_retry_number' => (int) $this->smsapiSettings['retry_number']], $this->uid, $this->userData);

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function getHelpLinks() {
    // @phpstan-ignore-next-line
    return ($this->pluginDefinition['helpLinks']) ?: '';
  }

  /**
   * {@inheritdoc}
   */
  public function getSetupMessages() {
    // @phpstan-ignore-next-line
    return ($this->pluginDefinition['setupMessages']) ?: '';
  }

  /**
   * {@inheritdoc}
   */
  public function getOverview(array $params) {
    /** @var \Drupal\Core\StringTranslation\TranslatableMarkup $label */
    $label = $this->getLabel();

    $plugin_text = $this->t('Validation Plugin: @plugin',
      [
        '@plugin' => str_replace(' Setup', '', $label->render()),
      ]
    );

    $output = [
      'heading' => [
        '#type' => 'html_tag',
        '#tag' => 'h2',
        '#value' => $this->t('SMS API verification code'),
      ],
      'validation_plugin' => [
        '#type' => 'markup',
        '#markup' => '<p>' . $plugin_text . '</p>',
      ],
      'description' => [
        '#type' => 'html_tag',
        '#tag' => 'p',
        '#value' => $this->t('Generate verification code from SMS API service.'),
      ],
      'link' => [
        '#theme' => 'links',
        '#links' => [
          'admin' => [
            'title' => !$params['enabled'] ? $this->t('Set up application') : $this->t('Reset application'),
            'url' => Url::fromRoute('tfa.validation.setup', [
              'user' => $params['account']->id(),
              'method' => $params['plugin_id'],
            ]),
          ],
        ],
      ],
    ];

    return $output;
  }

}
