<?php

declare(strict_types=1);

namespace Drupal\tfa_smsapi\Plugin\TfaValidation;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Utility\Crypt;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\HtmlCommand;
use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Lock\LockBackendInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\encrypt\EncryptionProfileManagerInterface;
use Drupal\encrypt\EncryptServiceInterface;
use Drupal\smsapi\Services\SmsapiService;
use Drupal\tfa\Plugin\TfaBasePlugin;
use Drupal\tfa\Plugin\TfaValidationInterface;
use Drupal\user\UserDataInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * SMS API verification class for performing SMS API validation.
 *
 * @TfaValidation(
 *   id = "tfa_smsapi",
 *   label = @Translation("TFA SMSAPI verification"),
 *   description = @Translation("TFA SMSAPI verification Plugin"),
 *   setupPluginId = "tfa_smsapi_setup",
 * )
 */
class TfaSmsapiValidation extends TfaBasePlugin implements TfaValidationInterface, ContainerFactoryPluginInterface {

  use StringTranslationTrait;

  /**
   * The default sender name for SMS messages.
   */
  const DEFAULT_SENDER_NAME = 'Test';

  /**
   * Authentication SMS API settings.
   */
  protected array $smsapiSettings;

  /**
   * Phone number.
   */
  protected ?string $phone;

  /**
   * Error message.
   */
  protected string $errorMessage;

  /**
   * Constructs a new Tfa plugin object.
   *
   * @param array $configuration
   *   The plugin configuration.
   * @param string $plugin_id
   *   The plugin id.
   * @param array $plugin_definition
   *   The plugin definition.
   * @param \Drupal\user\UserDataInterface $user_data
   *   User data object to store user specific information.
   * @param \Drupal\encrypt\EncryptionProfileManagerInterface $encryption_profile_manager
   *   Encryption profile manager.
   * @param \Drupal\encrypt\EncryptServiceInterface $encrypt_service
   *   Encryption service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration factory.
   * @param \Drupal\smsapi\Services\SmsapiService $smsapi_service
   *   The SMS API service.
   * @param \Symfony\Component\HttpFoundation\RequestStack $requestStack
   *   The request.
   * @param \Drupal\Core\Database\Connection $connection
   *   The connection service.
   * @param \Drupal\Core\Lock\LockBackendInterface $lock
   *   The lock service.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    array $plugin_definition,
    protected UserDataInterface $user_data,
    EncryptionProfileManagerInterface $encryption_profile_manager,
    EncryptServiceInterface $encrypt_service,
    protected ConfigFactoryInterface $config_factory,
    protected SmsApiService $smsapi_service,
    protected RequestStack $requestStack,
    protected Connection $connection,
    protected LockBackendInterface $lock,
    protected TimeInterface $time,
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $user_data, $encryption_profile_manager, $encrypt_service);
    $pluginSettings = $config_factory->get('tfa.settings')->get('validation_plugin_settings');
    $settings = $pluginSettings['tfa_smsapi'] ?? [];
    $settings = array_replace([
      'sender_name' => self::DEFAULT_SENDER_NAME,
      'retry_number' => 3,
    ], $settings);
    $this->smsapiSettings = $settings;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('user.data'),
      $container->get('encrypt.encryption_profile.manager'),
      $container->get('encryption'),
      $container->get('config.factory'),
      $container->get('smsapi.service'),
      $container->get('request_stack'),
      $container->get('database'),
      $container->get('lock'),
      $container->get('datetime.time'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getForm(array $form, FormStateInterface $form_state) {
    $message = $this->t('Verification code is SMSAPI generated and 6 digits long.');

    if (!empty($this->getUserData('tfa', 'tfa_recovery_code', $this->uid, $this->userData))) {
      $message .= '<br/>' . $this->t("Can't access your account? Use one of your recovery codes.");
    }

    $restricted = $this->checkForAccess($form);

    if (is_array($restricted)) {
      return $restricted;
    }

    $form['country_code'] = [
      '#type' => 'hidden',
    ];

    $form['phone'] = [
      '#type' => 'hidden',
    ];

    $form['user'] = [
      '#type' => 'hidden',
      '#value' => $this->uid,
    ];

    $form['get_code'] = [
      '#type' => 'button',
      '#value' => $this->t('Get verification code'),
      '#description' => $this->t('Click this button to recieve a SMS with the verification code.'),
      '#ajax' => [
        'callback' => 'Drupal\tfa_smsapi\Plugin\TfaValidation\TfaSmsapiValidation::getCode',
        'event' => 'click',
        'wrapper' => 'message-container',
      ],
    ];

    $form['message'] = [
      '#type' => 'markup',
      '#prefix' => '<div id="message-container">',
      '#suffix' => '</div>',
    ];

    $form['code'] = [
      '#type' => 'textfield',
      '#title' => $this->t('SMS verification code'),
      '#description' => $message,
      '#required'  => TRUE,
      '#attributes' => ['autocomplete' => 'off'],
    ];

    $form['#cache'] = ['max-age' => 0];
    $form['actions']['#type'] = 'actions';
    $form['actions']['login'] = [
      '#type' => 'submit',
      '#button_type' => 'primary',
      '#value' => $this->t('Verify'),
    ];

    return $form;
  }

  /**
   * Sends the code to the users mobile phone.
   *
   * @param array &$form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state object.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   Returns ajax message.
   */
  public static function getCode(array &$form, FormStateInterface $form_state): AjaxResponse {
    $ajaxResponse = new AjaxResponse();

    $phone = \Drupal::service('user.data')->get('tfa', $form_state->getValue('user'), 'tfa_smsapi_phone');

    if (empty($phone)) {
      $phone = $form_state->getValue('country_code') . $form_state->getValue('phone');
    }
    else {
      $phone = str_replace('-', '', $phone);
    }

    $senderName = \Drupal::config('tfa.settings')->get('validation_plugin_settings.tfa_smsapi')['sender_name'];

    if (empty($senderName)) {
      $senderName = self::DEFAULT_SENDER_NAME;
    }

    $response = \Drupal::service('smsapi.service')->sendVerificationCode($phone, '', [], $senderName);

    if ($response !== NULL) {
      $message = new TranslatableMarkup('Verification code has been sent via SMS to your phone. Please enter it below.');
    }
    else {
      $message = new TranslatableMarkup('There was an error while sending SMS. Please contact your site administrator.');
    }

    $ajaxResponse->addCommand(new HtmlCommand('#message-container', $message->render()));

    return $ajaxResponse;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array $form, FormStateInterface $form_state) {
    $code = $form_state->getValue('code');
    $phone = $this->getUserData('tfa', 'tfa_smsapi_phone', $this->uid, $this->userData);

    if (empty($phone)) {
      $phone = $form_state->getValue('country_code') . '-' . $form_state->getValue('phone');
    }

    $this->phone = $phone;

    /** @var \Symfony\Component\HttpFoundation\Request $request */
    $request = $this->requestStack->getCurrentRequest();

    // Don't validate if it's an ajax request.
    if ($request->isXmlHttpRequest()) {
      $form_state->clearErrors();
      return TRUE;
    }

    if (empty($code)) {
      $form_state->clearErrors();
      return FALSE;
    }

    $smsapiValidationLockId = 'tfa_validation_smsapi_' . $this->uid;

    while (!$this->lock->acquire($smsapiValidationLockId)) {
      $this->lock->wait($smsapiValidationLockId);
    }

    if (!$this->validate($code)) {
      $this->lock->release($smsapiValidationLockId);

      if ($this->alreadyAccepted) {
        $form_state->clearErrors();
        $this->errorMessages['code'] = $this->t('Invalid code, it was recently used for a login. Please try a new code.') . ' ' . $this->getAttemptNumberMessage();
      }
      else {
        $form_state->clearErrors();
        $attempts = $this->getAttemptNumberMessage() ?? '';
        $this->errorMessages['code'] = $this->errorMessage . '. ' . $attempts;
      }

      return FALSE;
    }
    else {
      $form_state->clearErrors();
      $this->storeAcceptedCode($code);
      $this->lock->release($smsapiValidationLockId);

      return TRUE;
    }
  }

  /**
   * Checks if user can access this tfa method.
   *
   * @param array $form
   *   Form element array.
   * @param bool $connectionOnly
   *   If we should check connection only.
   *
   * @return array|null
   *   Array on restriction or null if accessible.
   */
  protected function checkForAccess(array &$form, bool $connectionOnly = FALSE): ?array {
    $connected = $this->smsapi_service->checkIfConnected();

    if (!$connected) {
      $this->errorMessages['code'] = $this->t('Could not connect to SMSAPI, check your token.')->render();

      $form['blocked'] = [
        '#type' => 'markup',
        '#markup' => '<div class="readonly-text">' . $this->errorMessages['code'] . '</div>',
      ];

      return $form;
    }

    $data = $this->getUserData('tfa', 'tfa_smsapi_retry_number', $this->uid, $this->userData);

    if (!$connectionOnly && $data !== NULL && (int) $data <= 0) {
      $this->errorMessages['code'] = $this->t('Too many unsuccessful login attempts, you have been blocked. Please contact your site administrator.')->render();

      $form['blocked'] = [
        '#type' => 'markup',
        '#markup' => '<div class="readonly-text">' . $this->errorMessages['code'] . '</div>',
      ];

      return $form;
    }

    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  protected function storeAcceptedCode($code): void {
    /** @var string $code */
    $code = preg_replace('/\s+/', '', $code);
    $hash = Crypt::hashBase64($code);

    $this->deleteOldCodes();

    // Store the hash made using the code in users_data.
    $store_data = ['tfa_smsapi_accepted_code_' . $hash => $this->time->getRequestTime()];
    $this->setUserData('tfa', $store_data, $this->uid, $this->userData);
    // Renew number of retries.
    $this->setUserData('tfa', ['tfa_smsapi_retry_number' => (int) $this->smsapiSettings['retry_number']], $this->uid, $this->userData);
  }

  /**
   * Deletes old codes, we want to store only the latest one.
   */
  protected function deleteOldCodes(): void {
    $this->connection->delete('users_data')
      ->condition('uid', $this->uid)
      ->condition('module', 'tfa')
      ->condition('name', '%' . $this->connection->escapeLike('tfa_smsapi_accepted_code_') . '%', 'LIKE')
      ->execute();
  }

  /**
   * {@inheritdoc}
   */
  protected function alreadyAcceptedCode($code) {
    $hash = Crypt::hashBase64($code);
    // Check if the code has already been used or not.
    $key = 'tfa_smsapi_accepted_code_' . $hash;
    $result = $this->getUserData('tfa', $key, $this->uid, $this->userData);
    if (!empty($result)) {
      $this->alreadyAccepted = TRUE;
      return TRUE;
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  protected function validate($code) {
    if ($this->alreadyAcceptedCode($code)) {
      $this->decrementRetries();
      $this->isValid = FALSE;
    }
    else {
      $phone = str_replace('-', '', $this->phone ?? '');
      $result = $this->smsapi_service->checkVerificationCode($code, $phone);

      if ($result['status'] === TRUE) {
        $this->setUserData('tfa', ['tfa_smsapi_phone' => $this->phone], $this->uid, $this->userData);
        $this->isValid = TRUE;
      }
      else {
        $this->decrementRetries();
        $this->errorMessage = $result['message'];
        $this->isValid = FALSE;
      }
    }

    return $this->isValid;
  }

  /**
   * Returns the message of number of attempts left for a user.
   */
  protected function getAttemptNumberMessage(): ?TranslatableMarkup {
    $data = $this->getUserData('tfa', 'tfa_smsapi_retry_number', $this->uid, $this->userData);

    if (isset($data)) {
      return $this->t('You have @num attempts left.', ['@num' => (int) $data]);
    }
    else {
      return NULL;
    }
  }

  /**
   * Decrements number of allowed retries for a user.
   */
  protected function decrementRetries(): void {
    $data = $this->getUserData('tfa', 'tfa_smsapi_retry_number', $this->uid, $this->userData);

    if (isset($data)) {
      $retryNumber = (int) $data;

      $this->setUserData('tfa', ['tfa_smsapi_retry_number' => --$retryNumber], $this->uid, $this->userData);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function ready(): bool {
    $result = $this->getUserData('tfa', 'tfa_smsapi_phone', $this->uid, $this->userData);
    $connected = $this->smsapi_service->checkIfConnected();

    return (!empty($result) && $connected);
  }

  /**
   * The configuration form for SMS API plugin.
   *
   * @param \Drupal\Core\Config\Config $config
   *   Config object for tfa settings.
   * @param array $state
   *   Form state array determines if this form should be shown.
   *
   * @return array
   *   Form array specific for this validation plugin.
   */
  public function buildConfigurationForm(Config $config, array $state = []): array {
    $settingsForm = [];

    $restricted = $this->checkForAccess($settingsForm, TRUE);

    if (is_array($restricted)) {
      return $restricted;
    }

    $options = $this->smsapi_service->getSenders();

    if (array_key_exists('2WAY', $options)) {
      unset($options['2WAY']);
    }

    $settingsForm['sender_name'] = [
      '#type' => 'select',
      '#title' => $this->t('Please select the default SMS sender.'),
      '#description' => $this->t('This name will appear in the sender name of the SMS.'),
      '#options' => $options,
      '#default_value' => $this->smsapiSettings['sender_name'] ?? self::DEFAULT_SENDER_NAME,
    ];

    $settingsForm['retry_number'] = [
      '#type' => 'number',
      '#title' => $this->t('Enter a number of available retries.'),
      '#default_value' => $this->smsapiSettings['retry_number'] ?? 3,
      '#description' => $this->t('Specify how many times user will be able to retry after entering an invalid code until he is blocked.'),
      '#min' => 1,
      '#max' => 10,
      '#step' => 1,
    ];

    return $settingsForm;
  }

}
